"""External functions patched or extended for outset compatibility."""

from ._annotateplot import annotateplot
from ._regplot import regplot
from ._scatterplot import scatterplot

__all__ = [
    "annotateplot",
    "regplot",
    "scatterplot",
]



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("../src/"))

import outset


# -- Project information -----------------------------------------------------

project = "outset"
copyright = "2023, Matthew Andres Moreno"
author = "Matthew Andres Moreno"

# The version info for the project you're documenting, acts as replacement
# for |version| and |release|, also used in various other places throughout
# the built documents.
#
# The short X.Y version.
version = outset.__version__
# The full version, including alpha/beta/rc tags.
release = outset.__version__

autosummary_generate = True
autosummary_imported_members = True

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "sphinx.ext.ifconfig",
    "sphinx.ext.viewcode",  # Add links to highlighted source code
    "sphinx.ext.napoleon",  # to render Google format docstrings
    "sphinx.ext.githubpages",
    "sphinx_rtd_theme",
    "nbsphinx",
]

autoclass_content = "class"

autodoc_class_signature = "separated"

add_module_names = False

autodoc_default_options = {
    "member-order": "bysource",
    "special-members": "__init__",
    "undoc-members": True,
    "exclude-members": "__weakref__",
}

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]


# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_js_files = [
    "hide_code_cells.js",
]


def setup(app):
    app.add_css_file("theme_override.css")


# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "assets/outset-wordmark.png"


# Napoleon settings
napoleon_include_init_with_doc = True
napoleon_include_private_with_doc = True



from matplotlib import pyplot as plt
import pandas as pd
import seaborn as sns


def scatterplot(
    data: pd.DataFrame,
    *args: list,
    **kwargs: dict,
) -> plt.Axes:
    """Wrapper around sns.scatterplot patching seaborn issue #3601.

    See https://github.com/mwaskom/seaborn/issues/3601.
    """

    filter = data.index == data.index
    if "hue" in kwargs and "hue_order" in kwargs:
        filter &= data[kwargs["hue"]].isin(kwargs["hue_order"])

    if "style" in kwargs and "style_order" in kwargs:
        filter &= data[kwargs["style"]].isin(kwargs["style_order"])

    data = data[filter].reset_index()

    return sns.scatterplot(data, *args, **kwargs)



import itertools as it
import math
import typing

import numpy as np


def layout_corner_insets(
    num_insets: int,
    corner: typing.Literal["NE", "NW", "SE", "SW"] = "NE",
    *,
    inset_grid_size: typing.Union[typing.Tuple[float, float], float] = 0.50,
    inset_margin_size: typing.Union[typing.Tuple[float, float], float] = 0.1,
    inset_pad_ratio: typing.Union[typing.Tuple[float, float], float] = 0.1,
    transpose: bool = False,
) -> typing.List[typing.Tuple[float, float, float, float]]:
    """Lay out positions for `n` inset plots in a specified corner.

    The frames are filled from the chosen corner outwards, filling successive diagonals perpendicular to the cornder. Inset subfigure positions are
    ordered so that inset number increases left-to-right, top-to-bottom.

    Parameters
    ----------
    num_insets : int
        The number of inset plots to be generated.
    corner : Literal["NE", "NW", "SE", "SW"], default "NE"
        The corner of the grid where the insets will be positioned.
    inset_margin_size : Union[Tuple[float, float], float], default 0.5
        How far should grid be spaced from source plot boundaries, relative to
        the source plot size?
    inset_grid_size : Union[Tuple[float, float], float], default 0.5
        The size of the grid of inset plots relative to the source plot.
    inset_pad_ratio : Union[Tuple[float, float], float], default 0.1
        Pad size between inset plots as a fraction of inset plot size.
    transpose : bool, default False
        Should inset grid layout be ordered top-to-bottom, left-to-right?

    Returns
    -------
    List[Tuple[float, float, float, float]]
        A list of tuples, each representing `(x, y, width, height)` of an inset
        plot fractionally relative to source plot axes.

    See Also
    --------
    outset.util.inset_outsets
        Manipulates plot structure to apply corner inset layout calculated by
        `layout_corner_insets`, placing outset plots in specified positions over
        the source plot axes.
    """
    dimension = int(math.ceil(math.sqrt(num_insets)))
    if not isinstance(inset_grid_size, typing.Iterable):
        inset_grid_size = (inset_grid_size, inset_grid_size)

    if not np.allclose(np.clip(inset_grid_size, 0.0, 1.0), inset_grid_size):
        raise ValueError("inset_grid_size values must be within unit scale")

    grid_size_x = inset_grid_size[0] / dimension
    grid_size_y = inset_grid_size[1] / dimension

    if not isinstance(inset_pad_ratio, typing.Iterable):
        inset_pad_ratio = (inset_pad_ratio, inset_pad_ratio)

    if not np.allclose(np.clip(inset_pad_ratio, 0.0, 1.0), inset_pad_ratio):
        raise ValueError("inset_pad_ratio values must be within unit scale")

    ax_size_x = grid_size_x * (1 - inset_pad_ratio[0])
    assert ax_size_x < grid_size_x

    ax_size_y = grid_size_y * (1 - inset_pad_ratio[1])
    assert ax_size_y < grid_size_y

    if not isinstance(inset_margin_size, typing.Iterable):
        inset_margin_size = (inset_margin_size, inset_margin_size)

    if not np.allclose(np.clip(inset_margin_size, 0.0, 1.0), inset_margin_size):
        raise ValueError("inset_margin_size values must be within unit scale")
    margin_x, margin_y = inset_margin_size

    if margin_x + grid_size_x > 1:
        raise ValueError(
            f"x grid size {grid_size_x} with margin {margin_x} "
            "exceeds available space"
        )
    if margin_y + grid_size_y > 1:
        raise ValueError(
            f"y grid size {grid_size_y} with margin {margin_y} "
            "exceeds available space"
        )

    # generate frame coordinates from the corner out, along diagonals
    # https://mmore500.com/2023/12/16/zigzag-traversal.html
    frame_coordinates = []
    for row, col in sorted(
        sorted(it.product(range(dimension), repeat=2), key=sum)[:num_insets],
    ):
        if transpose:
            row, col = col, row
        # Calculate the bottom-left corner of each frame
        x = 1.0 - margin_x - ax_size_x - col * grid_size_x
        y = 1.0 - margin_y - ax_size_y - row * grid_size_y
        assert 1 - inset_grid_size[0] - margin_x <= x + 1e-6
        assert x <= 1.0 - margin_x + 1e-6
        assert 1 - inset_grid_size[1] - margin_y <= y + 1e-6
        assert y <= 1.0 - margin_y + 1e-6

        if corner == "NW" or corner == "SW":
            x = grid_size_x * col + margin_x
            assert margin_x <= x + 1e-6
            assert x <= inset_grid_size[0] - ax_size_x + margin_x + 1e-6
        if corner == "SE" or corner == "SW":
            y = grid_size_y * row + margin_y
            assert margin_y <= y + 1e-6
            assert y <= inset_grid_size[1] - ax_size_y + margin_y + 1e-6

        # Add the frame to the list
        frame_coordinates.append((x, y, ax_size_x, ax_size_y))

    # take the first `num_insets` frames from corner out and then sort them
    # so that frames are ordered left-to-right, top-to-bottom
    return sorted(frame_coordinates[:num_insets], key=lambda c: (-c[1], c[0]))



from matplotlib import pyplot as plt
import pytest

from outset._auxlib.calc_aspect_ import calc_aspect


def test_calc_aspect_square_physical():
    fig, ax = plt.subplots(figsize=(6, 6))  # Create a 6x6 inch figure
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    assert calc_aspect(ax) == pytest.approx(1, abs=0.1)


def test_calc_aspect_tall_physical():
    fig, ax = plt.subplots(figsize=(6, 12))  # Create a 6x12 inch figure
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    assert calc_aspect(ax) == pytest.approx(2, abs=0.1)


def test_calc_aspect_wide_physical():
    fig, ax = plt.subplots(figsize=(12, 6))  # Create a 12x6 inch figure
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    assert calc_aspect(ax) == pytest.approx(0.5, abs=0.1)


def test_calc_aspect_square():
    fig, ax = plt.subplots(figsize=(6, 6))  # Create a 6x6 inch figure
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 1)
    assert calc_aspect(ax) == pytest.approx(1, abs=0.1)


def test_calc_aspect_tall():
    fig, ax = plt.subplots(figsize=(6, 6))  # Create a 6x6 inch figure
    ax.set_xlim(0, 1)
    ax.set_ylim(0, 2)
    assert calc_aspect(ax) == pytest.approx(0.5, abs=0.1)


def test_calc_aspect_wide():
    fig, ax = plt.subplots(figsize=(6, 6))  # Create a 6x6 inch figure
    ax.set_xlim(-1, 3)
    ax.set_ylim(0, 2)
    assert calc_aspect(ax) == pytest.approx(2, abs=0.1)



import seaborn as sns
from matplotlib.testing.decorators import check_figures_equal

from outset.patched import scatterplot


@check_figures_equal(extensions=["png"])
def test_scatterplot_without_hue(fig_test, fig_ref):
    """Test scatterplot without a hue parameter."""
    # Create the test figure
    ax_test = fig_test.subplots()
    scatterplot(
        data=sns.load_dataset("tips"), x="total_bill", y="tip", ax=ax_test
    )

    # Create the reference figure
    ax_ref = fig_ref.subplots()
    sns.scatterplot(
        data=sns.load_dataset("tips"), x="total_bill", y="tip", ax=ax_ref
    )


@check_figures_equal(extensions=["png"])
def test_scatterplot_with_hue(fig_test, fig_ref):
    """Test scatterplot with hue parameter."""
    # Create the test figure
    ax_test = fig_test.subplots()
    scatterplot(
        data=sns.load_dataset("tips"),
        x="total_bill",
        y="tip",
        hue="day",
        ax=ax_test,
    )

    # Create the reference figure
    ax_ref = fig_ref.subplots()
    sns.scatterplot(
        data=sns.load_dataset("tips"),
        x="total_bill",
        y="tip",
        hue="day",
        ax=ax_ref,
    )


@check_figures_equal(extensions=["png"])
def test_scatterplot_with_palette(fig_test, fig_ref):
    # Create the test figure
    ax_test = fig_test.subplots()
    scatterplot(
        data=sns.load_dataset("tips"),
        x="total_bill",
        y="tip",
        hue="time",
        palette="coolwarm",
        ax=ax_test,
    )

    # Create the reference figure
    ax_ref = fig_ref.subplots()
    sns.scatterplot(
        data=sns.load_dataset("tips"),
        x="total_bill",
        y="tip",
        hue="time",
        palette="coolwarm",
        ax=ax_ref,
    )


# Load an example dataset from seaborn
diamonds = sns.load_dataset("diamonds").dropna()


@check_figures_equal(extensions=["png"])
def test_scatterplot_hue_order_subset(fig_test, fig_ref):
    hue_order_subset = ["I1", "IF"]

    ax_test = fig_test.subplots()
    scatterplot(
        data=diamonds,
        x="price",
        y="carat",
        hue="clarity",
        hue_order=hue_order_subset,
        ax=ax_test,
    )

    ax_ref = fig_ref.subplots()
    sns.scatterplot(
        data=diamonds[diamonds["clarity"].isin(hue_order_subset)],
        x="price",
        y="carat",
        hue="clarity",
        hue_order=hue_order_subset,
        ax=ax_ref,
    )


@check_figures_equal(extensions=["png"])
def test_scatterplot_style_order_subset(fig_test, fig_ref):
    style_order_subset = ["I1", "IF"]

    ax_test = fig_test.subplots()
    scatterplot(
        data=diamonds,
        x="price",
        y="carat",
        style="clarity",
        style_order=style_order_subset,
        ax=ax_test,
    )

    ax_ref = fig_ref.subplots()
    sns.scatterplot(
        data=diamonds[diamonds["clarity"].isin(style_order_subset)],
        x="price",
        y="carat",
        style="clarity",
        style_order=style_order_subset,
        ax=ax_ref,
    )



import numpy as np
import pytest

from outset.util import layout_corner_insets


@pytest.mark.parametrize("transpose", [False, True])
@pytest.mark.parametrize(
    "num_insets, corner, expected_result",
    [
        (1, "NW", [(0.10, 0.70, 0.20, 0.20)]),
        (
            3,
            "NE",
            [
                (0.60, 0.80, 0.10, 0.10),
                (0.80, 0.80, 0.10, 0.10),
                (0.80, 0.60, 0.10, 0.10),
            ],
        ),
        (
            2,
            "SW",
            [
                (0.10, 0.10, 0.10, 0.10),
                (0.30, 0.10, 0.10, 0.10),
            ],
        ),
        (
            4,
            "SE",
            [
                (0.60, 0.30, 0.10, 0.10),
                (0.80, 0.30, 0.10, 0.10),
                (0.60, 0.10, 0.10, 0.10),
                (0.80, 0.10, 0.10, 0.10),
            ],
        ),
    ],
)
def test_layout_corner_insets(num_insets, corner, expected_result, transpose):
    result = layout_corner_insets(
        num_insets,
        corner,
        inset_grid_size=0.40,
        inset_margin_size=0.10,
        inset_pad_ratio=0.50,
        transpose=transpose,
    )
    if transpose and corner in ("NE", "SW"):
        expected_result = [
            (*reversed(coord[:2:]), *coord[2:]) for coord in expected_result
        ]
        expected_result = [*reversed(expected_result)]
    elif transpose:
        pass  # too complicated, not worth testing directly
    assert np.allclose(result, expected_result)


@pytest.mark.parametrize(
    "num_insets, corner, expected_result",
    [
        (1, "NW", [(0.10, 0.70, 0.20, 0.20)]),
        (
            3,
            "NE",
            [
                (0.60, 0.80, 0.10, 0.10),
                (0.80, 0.80, 0.10, 0.10),
                (0.80, 0.60, 0.10, 0.10),
            ],
        ),
        (
            2,
            "SW",
            [
                (0.10, 0.10, 0.10, 0.10),
                (0.30, 0.10, 0.10, 0.10),
            ],
        ),
        (
            4,
            "SE",
            [
                (0.60, 0.30, 0.10, 0.10),
                (0.80, 0.30, 0.10, 0.10),
                (0.60, 0.10, 0.10, 0.10),
                (0.80, 0.10, 0.10, 0.10),
            ],
        ),
    ],
)
def test_layout_corner_insets_tuple(num_insets, corner, expected_result):
    result = layout_corner_insets(
        num_insets,
        corner,
        inset_grid_size=(0.40, 0.40),
        inset_margin_size=(0.10, 0.10),
        inset_pad_ratio=(0.50, 0.50),
    )
    assert np.allclose(result, expected_result)



import numbers
import typing
import warnings

import numpy as np
from matplotlib import axes as mpl_axes

# for several implementations, we need to calculate the padding size that will
# create the correct proportions AFTER padding has been set
# here's the math:
# because pad_axfrac * end_axdatawidth = end_paddatawidth
# so pad_axfrac = end_paddatawidth / end_axdatawidth
# and end_axdatawidth = start_axdatawidth + end_paddatawidth * 2
# so ...
# pad_axfrac = end_paddatawidth / (start_axdatawidth + end_paddatawidth * 2)
# paf = epdw / (sadw + epdw * 2)
# paf = 1 / (sadw / epdw + 2)
# paf * (sadw / epdw + 2) = 1
# paf * sadw / epdw + paf * 2 = 1
# paf * sadw / epdw = 1 - paf * 2
# paf * sadw = epdw * (1 - paf * 2)
# which provides
# epdw = paf * sadw / (1 - paf * 2)


def _calc_pad_x(ax: mpl_axes.Axes, x_axfrac: float) -> float:
    if x_axfrac >= 0.5:
        warnings.warn("x outer pad exceeds half axis size, clipping to 0.5")
        x_axfrac = 0.5
    pad_x = x_axfrac * np.ptp(ax.get_xlim()) / (1 - x_axfrac * 2)
    assert np.isclose((np.ptp(ax.get_xlim()) + 2 * pad_x) * x_axfrac, pad_x)
    assert np.isfinite(pad_x)
    return pad_x


def _calc_pad_y(ax: mpl_axes.Axes, y_axfrac: float) -> float:
    if y_axfrac >= 0.5:
        warnings.warn("x outer pad exceeds half axis size, clipping to 0.5")
        y_axfrac = 0.5
    pad_y = y_axfrac * np.ptp(ax.get_ylim()) / (1 - y_axfrac * 2)
    assert np.isclose((np.ptp(ax.get_ylim()) + 2 * pad_y) * y_axfrac, pad_y)
    assert np.isfinite(pad_y)
    return pad_y


def _calc_outer_pad_axes(
    ax: mpl_axes.Axes, frame_outer_pad: float
) -> typing.Tuple[float, float]:
    """Calculate width in data units necessary for pad to be a fraction
    `frame_outer_pad` of axes size."""
    return _calc_pad_x(ax, frame_outer_pad), _calc_pad_y(ax, frame_outer_pad)


def _calc_outer_pad_figure(
    ax: mpl_axes.Axes, frame_outer_pad: float
) -> typing.Tuple[float, float]:
    """Calculate width in data units necessary for pad to be a fraction
    `frame_outer_pad` of figure size."""
    if frame_outer_pad >= 0.5:
        raise ValueError(
            "frame_outer_pad must be less than 0.5 when "
            "frame_outer_pad_unit='figure'",
        )
    pad_x_axfrac = frame_outer_pad / ax.get_position().width
    pad_y_axfrac = frame_outer_pad / ax.get_position().height
    return _calc_pad_x(ax, pad_x_axfrac), _calc_pad_y(ax, pad_y_axfrac)


def _calc_outer_pad_inches(
    ax: mpl_axes.Axes, frame_outer_pad: float
) -> typing.Tuple[float, float]:
    """Calculate width in data units necessary for pad to be rendered as `frame_outer_pad` inches."""
    axwidth_inches = ax.get_position().width * ax.figure.get_figwidth()
    axheight_inches = ax.get_position().height * ax.figure.get_figheight()
    pad_x_axfrac = frame_outer_pad / axwidth_inches
    pad_y_axfrac = frame_outer_pad / axheight_inches
    return _calc_pad_x(ax, pad_x_axfrac), _calc_pad_y(ax, pad_y_axfrac)


def calc_outer_pad(
    ax: mpl_axes.Axes, frame_outer_pad: float, frame_outer_pad_unit: str
) -> typing.Tuple[float, float]:
    """Calculate pad width in data units."""
    if isinstance(frame_outer_pad, numbers.Number):
        try:
            return {
                "axes": _calc_outer_pad_axes,
                "figure": _calc_outer_pad_figure,
                "inches": _calc_outer_pad_inches,
            }[frame_outer_pad_unit](ax, frame_outer_pad)
        except KeyError:
            raise ValueError(
                "frame_outer_pad_unit must be 'axes', 'figure', or 'inches',"
                f"not '{frame_outer_pad_unit}'",
            )
    else:
        return frame_outer_pad



import pandas as pd
import matplotlib.pyplot as plt
import pytest

from outset import patched as otst_patched


@pytest.fixture
def sample_data():
    """Provide a small dataset for testing."""
    return pd.DataFrame(
        {"x": [1, 2, 3], "y": [4, 5, 6], "text": ["A", "B", "C"]}
    )


def test_annotateplot(sample_data: pd.DataFrame):
    """Test the annotateplot function with minimal input."""
    fig, ax = plt.subplots()
    result = otst_patched.annotateplot(
        data=sample_data, x="x", y="y", text="text", ax=ax
    )
    assert isinstance(result, plt.Axes)

    outpath = "/tmp/test_annotateplot.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import matplotlib.pyplot as plt
import pandas as pd
import pytest
import seaborn as sns

from outset import inset_outsets, OutsetGrid

# Sample data for testing
data = pd.DataFrame(
    {
        "x": [1, 10, 1.5, 4, 7, 7.4, 4, 5],
        "y": [1.5, 3, 2, 11, 2.05, 2, 4, 5],
        "outset": ["A", "B", "A", "B", "C", "C", "D", "D"],
    }
)


def test_inset_outsets_one():
    # Create a sample dataframe
    og = OutsetGrid(
        data=data,
        x="x",
        y="y",
        col="outset",
        col_order=["A"],
        marqueeplot_kws={
            "frame_inner_pad": 0.1,
            "frame_outer_pad": 0.3,
        },
    )
    og.map_dataframe(sns.scatterplot, x="x", y="y")
    og.marqueeplot()
    inset_outsets(og, [(0.8, 0.6, 0.1, 0.1)])
    outpath = "/tmp/test_inset_outsets_one.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_inset_outsets_two():
    # Create a sample dataframe
    og = OutsetGrid(
        data=data,
        x="x",
        y="y",
        col="outset",
        col_order=["A", "C"],
        marqueeplot_kws={
            "frame_inner_pad": 0.3,
            "frame_outer_pad": 0.1,
            "frame_outer_pad_unit": "axes",
        },
        marqueeplot_outset_kws={
            "frame_outer_pad": 0.2,
            "frame_outer_pad_unit": "inches",
            "mark_glyph_kws": {"markersize": 16},
        },
        aspect=1.2,
    )
    og.map_dataframe(sns.scatterplot, x="x", y="y", hue="outset", legend=False)
    inset_outsets(
        og,
        [
            (0.1, 0.5, 0.3, 0.3),
            (0.5, 0.5, 0.5, 0.3),
        ],
    )
    og.marqueeplot()

    outpath = "/tmp/test_inset_outsets_two.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


@pytest.mark.parametrize("corner", ["NE", "NW", "SE", "SW"])
def test_inset_outsets_three(corner: str):
    # Create a sample dataframe
    og = OutsetGrid(
        data=data,
        x="x",
        y="y",
        col="outset",
        col_order=["A", "B", "C"],
        marqueeplot_kws={
            "frame_inner_pad": 0.1,
            "frame_outer_pad": 0.1,
            "frame_outer_pad_unit": "axes",
        },
        marqueeplot_outset_kws={
            "frame_outer_pad": 0.1,
            "frame_outer_pad_unit": "inches",
            "mark_glyph_kws": {"markersize": 16},
        },
        aspect=1.2,
    )
    og.map_dataframe(
        sns.scatterplot, x="x", y="y", hue="outset", legend=False, zorder=-1
    )
    inset_outsets(og, insets=corner, strip_spines=False)
    og.equalize_aspect()
    og.marqueeplot()

    assert og._is_inset()

    outpath = f"/tmp/test_inset_outsets_three_{corner}.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import matplotlib.cbook as mpl_cbook
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns

from outset import inset_outsets, OutsetGrid
from outset import util as otst_util

# Sample data for testing
data = pd.DataFrame(
    {"x": [1, 2, 3, 4], "y": [1, 3, 2, 1], "outset": ["A", "B", "A", "B"]}
)


def test_OutsetGrid_one():
    # Create a sample dataframe
    data = pd.DataFrame({"x": [0.5], "y": [1], "outset": ["A"]})
    sns.scatterplot(data=data, x="x", y="y")
    g = OutsetGrid(
        data=data,
        x="x",
        y="y",
        marqueeplot_kws={
            "frame_inner_pad": 0.2,
        },
    ).marqueeplot()

    assert not g._is_inset()

    outpath = "/tmp/test_OutsetGrid_one.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_with_sourceplot_monochrome():
    # Create sample data with a 'outset' column for grouping
    data = pd.DataFrame(
        {
            "x": [0.825, 3.1, 0.5, 0.8, 2.2, 2],
            "y": [1.2, 0.8, 2.5, 2.3, 1.1, 3.7],
            "outset": [
                "group1",
                "group1",
                "group2",
                "group2",
                "group3",
                "group3",
            ],
        }
    )

    g = OutsetGrid(data=data, x="x", y="y", col="outset")
    g.marqueeplot()
    g.map_dataframe(sns.scatterplot, x="x", y="y", hue="outset", legend=False)

    assert not g._is_inset()

    outpath = "/tmp/test_OutsetGrid_with_sourceplot_monochrome.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_with_sourceplot_hue():
    # Create sample data with a 'outset' column for grouping
    data = pd.DataFrame(
        {
            "x": [0.825, 3.1, 0.5, 0.8, 2.2, 2],
            "y": [1.2, 0.8, 2.5, 2.3, 1.1, 3.7],
            "outset": [
                "group1",
                "group1",
                "group2",
                "group2",
                "group3",
                "group3",
            ],
        }
    )

    g = OutsetGrid(data=data, x="x", y="y", hue="outset")
    g.marqueeplot()
    g.map_dataframe(
        sns.scatterplot,
        x="x",
        y="y",
        marker=otst_util.SplitKwarg(outset="+", source="o"),
        legend=False,
    )

    assert not g._is_inset()

    outpath = "/tmp/test_OutsetGrid_with_sourceplot_hue.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_broadcast():
    with mpl_cbook.get_sample_data("grace_hopper.jpg") as image_file:
        image = plt.imread(image_file)

    og = OutsetGrid(
        data=[(0.42, 0.78, 0.62, 0.98), (0.10, 0.14, 0.40, 0.21)],
        aspect=0.9,
    )
    og.broadcast(
        plt.imshow, image, extent=(0, 1, 0, 1), origin="upper", zorder=-1
    )
    og.broadcast(
        plt.plot,
        [0.5],
        [0.85],
        color="red",
        linewidth=2,
        marker=otst_util.SplitKwarg(outset="+", source="o"),
        zorder=10,
    )
    og.marqueeplot()

    assert not og._is_inset()

    outpath = "/tmp/test_OutsetGrid_broadcast.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_broadcast_sourceless():
    with mpl_cbook.get_sample_data("grace_hopper.jpg") as image_file:
        image = plt.imread(image_file)

    og = OutsetGrid(
        data=[(0.42, 0.78, 0.62, 0.98), (0.10, 0.14, 0.40, 0.21)],
        aspect=0.9,
        include_sourceplot=False,
    )
    og.broadcast(
        plt.imshow, image, extent=(0, 1, 0, 1), origin="upper", zorder=-1
    )
    og.marqueeplot(equalize_aspect=False, preserve_aspect=True)

    assert not og._is_inset()

    outpath = "/tmp/test_OutsetGrid_broadcast_sourceless.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_empty():
    og = OutsetGrid([])
    og.broadcast(sns.scatterplot, x=[1], y=[2], legend=False)


def test_OutsetGrid_with_row_singleton():
    # Create sample data with a 'outset' column for grouping
    data = pd.DataFrame(
        {
            "x": [0.825, 3.1, 0.5, 0.8, 2.2, 2],
            "y": [1.2, 0.8, 2.5, 2.3, 1.1, 3.7],
            "baz": [0, 0, 0, 1, 1, 1],
            "outset": [
                "group1",
                "group1",
                "group2",
                "group2",
                "group3",
                "group3",
            ],
        }
    )

    g = OutsetGrid(
        data=data, x="x", y="y", col="outset", row="baz", row_order=[0]
    )
    g.marqueeplot()
    g.map_dataframe(sns.scatterplot, x="x", y="y", hue="outset", legend=False)

    assert not g._is_inset()

    outpath = "/tmp/test_OutsetGrid_with_row_singleton.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_no_marqueeplot():
    grid = OutsetGrid(
        data=[[(-3, 0), (3, 4)]] * 4,
        aspect=1.5,
    )
    inset_outsets(
        grid,
        insets=otst_util.layout_corner_insets(
            1,
            "NW",
            inset_grid_size=0.35,
        )
        + otst_util.layout_corner_insets(
            3,
            "SE",
            inset_pad_ratio=0.4,
            inset_margin_size=(0.0, 0.1),
        ),
        equalize_aspect=False,
        strip_ticks=False,
    )

    grid.broadcast(
        plt.plot,
        np.linspace(-3, 3, 100),
        np.sin(np.linspace(-3, 3, 100)),
    )

    outpath = "/tmp/test_OutsetGrid_no_marqueeplot.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_OutsetGrid_n():
    grid = OutsetGrid(
        data=3,
        aspect=1.5,
        col_wrap=2,
    )

    outpath = "/tmp/test_OutsetGrid_n.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import matplotlib.pyplot as plt

from outset.mark import MarkRomanBadges


def test_MarkRomanBadges_lower():
    ax = plt.gca()  # Get the current axes

    ftor = MarkRomanBadges()
    # Call the function with test values
    ftor(x=0.5, y=0.5, ax=ax, color="red")
    ftor(
        x=1.5,
        y=1.5,
        ax=ax,
        scale_numeral=0.6,
    )

    # Save and print the path of the output image
    outpath = "/tmp/test_MarkRomanBadges_lower.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_MarkRomanBadges_upper():
    ax = plt.gca()  # Get the current axes

    ftor = MarkRomanBadges(upper=True)
    # Call the function with test values
    ftor(x=0.5, y=0.5, ax=ax, color="red")
    ftor(
        x=1.5,
        y=1.5,
        ax=ax,
        color="red",
        scale_numeral=0.6,
    )

    # Save and print the path of the output image
    outpath = "/tmp/test_MarkRomanBadges_upper.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")


def test_MarkRomanBadges_start():
    ax = plt.gca()  # Get the current axes

    ftor = MarkRomanBadges(start=9)
    # Call the function with test values
    ftor(x=0.5, y=0.5, ax=ax, color="red")
    ftor(
        x=1.5,
        y=1.5,
        ax=ax,
        color="red",
        scale_numeral=0.6,
    )

    # Save and print the path of the output image
    outpath = "/tmp/test_MarkRomanBadges_start.png"
    plt.savefig(outpath)
    print(f"saved graphic to {outpath}")



import math
import typing


def find_intersection(
    p1: typing.Tuple[float, float],
    p2: typing.Tuple[float, float],
    p3: typing.Tuple[float, float],
    p4: typing.Tuple[float, float],
) -> typing.Tuple[float, float]:
    """Finds the intersection of two lines (p1 to p2 and p3 to p4).

    Points are represented a tuple of (x, y).

    Returns
    -------
    Tuple[float, float]
        The intersection point, (x, y).

        If lines do not intersect or are underspecified, result is (nan, nan).
    """

    # Unpack points
    x1, y1 = p1
    x2, y2 = p2
    x3, y3 = p3
    x4, y4 = p4

    denominator = (x1 - x2) * (y3 - y4) - (y1 - y2) * (x3 - x4)
    if denominator == 0:
        return (math.nan, math.nan)

    # adapted from https://stackoverflow.com/a/51127674
    px = (
        (x1 * y2 - y1 * x2) * (x3 - x4) - (x1 - x2) * (x3 * y4 - y3 * x4)
    ) / denominator
    py = (
        (x1 * y2 - y1 * x2) * (y3 - y4) - (y1 - y2) * (x3 * y4 - y3 * x4)
    ) / denominator

    return (px, py)



import itertools as it
import typing

from adjustText import adjust_text
from frozendict import frozendict
import pandas as pd
from matplotlib import pyplot as plt


def annotateplot(
    data: pd.DataFrame,
    *,
    x: str,
    y: str,
    text: str,
    ax: typing.Optional[plt.Axes] = None,
    adjusttext_kws: typing.Mapping = frozendict(),
    **kwargs: dict,
) -> plt.Axes:
    """Annotate a plot coordinates with text labels, then apply adjustText to
    rearrange the labels to avoid overlaps.

    Parameters
    ----------
    data : pd.DataFrame
        The DataFrame containing the data to plot.
    x : str
        The name of the column in `data` to use for the x-axis values.
    y : str
        The name of the column in `data` to use for the y-axis values.
    text : Optional[str], default None
        The name of the column in `data` to use for text values.
    ax : Optional[plt.Axes], default None
        The matplotlib Axes object to draw the plot onto, if provided.
    adjusttext_kws : Mapping, default {}
        Additional keyword arguments forward to adjustText.
    **kwargs : dict
        Additional keyword arguments forward to seaborn's regplot.

    Returns
    -------
    plt.Axes
        The matplotlib Axes containing the plot.

    Notes
    -----
    This functionality is not provided by seaborn.
    """
    if ax is None:
        ax = plt.gca()

    kwargs.pop("legend", None)
    kwargs.pop("label", None)

    texts = [
        ax.text(row[x], row[y], row[text], **kwargs)
        for _idx, row in data.iterrows()
    ]
    adjust_text(texts, ax=ax, **adjusttext_kws)

    return ax
